@@ -18,6 +18,67 @@ import (
 	"testing"
 )
 
+func TestStreamWriteBlockedByOutputBuffer(t *testing.T) {
+	testStreamTypes(t, "", func(t *testing.T, styp streamType) {
+		ctx := canceledContext()
+		want := []byte{0, 1, 2, 3, 4, 5, 6, 7, 8, 9}
+		const writeBufferSize = 4
+		tc := newTestConn(t, clientSide, permissiveTransportParameters, func(c *Config) {
+			c.MaxStreamWriteBufferSize = writeBufferSize
+		})
+		tc.handshake()
+		tc.ignoreFrame(frameTypeAck)
+
+		s, err := tc.conn.newLocalStream(ctx, styp)
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		// Non-blocking write.
+		n, err := s.WriteContext(ctx, want)
+		if n != writeBufferSize || err != context.Canceled {
+			t.Fatalf("s.WriteContext() = %v, %v; want %v, context.Canceled", n, err, writeBufferSize)
+		}
+		tc.wantFrame("first write buffer of data sent",
+			packetType1RTT, debugFrameStream{
+				id:   s.id,
+				data: want[:writeBufferSize],
+			})
+		off := int64(writeBufferSize)
+
+		// Blocking write, which must wait for buffer space.
+		w := runAsync(tc, func(ctx context.Context) (int, error) {
+			return s.WriteContext(ctx, want[writeBufferSize:])
+		})
+		tc.wantIdle("write buffer is full, no more data can be sent")
+
+		// The peer's ack of the STREAM frame allows progress.
+		tc.writeAckForAll()
+		tc.wantFrame("second write buffer of data sent",
+			packetType1RTT, debugFrameStream{
+				id:   s.id,
+				off:  off,
+				data: want[off:][:writeBufferSize],
+			})
+		off += writeBufferSize
+		tc.wantIdle("write buffer is full, no more data can be sent")
+
+		// The peer's ack of the second STREAM frame allows sending the remaining data.
+		tc.writeAckForAll()
+		tc.wantFrame("remaining data sent",
+			packetType1RTT, debugFrameStream{
+				id:   s.id,
+				off:  off,
+				data: want[off:],
+			})
+
+		if n, err := w.result(); n != len(want)-writeBufferSize || err != nil {
+			t.Fatalf("s.WriteContext() = %v, %v; want %v, nil",
+				len(want)-writeBufferSize, err, writeBufferSize)
+		}
+	})
+}
+
 func TestStreamWriteBlockedByStreamFlowControl(t *testing.T) {
 	testStreamTypes(t, "", func(t *testing.T, styp streamType) {
 		ctx := canceledContext()
@@ -30,14 +91,15 @@ func TestStreamWriteBlockedByStreamFlowControl(t *testing.T) {
 		tc.handshake()
 		tc.ignoreFrame(frameTypeAck)
 
-		// Non-blocking write with no flow control.
 		s, err := tc.conn.newLocalStream(ctx, styp)
 		if err != nil {
 			t.Fatal(err)
 		}
-		_, err = s.WriteContext(ctx, want)
-		if err != context.Canceled {
-			t.Fatalf("write to stream with no flow control: err = %v, want context.Canceled", err)
+
+		// Data is written to the stream output buffer, but we have no flow control.
+		_, err = s.WriteContext(ctx, want[:1])
+		if err != nil {
+			t.Fatalf("write with available output buffer: unexpected error: %v", err)
 		}
 		tc.wantFrame("write blocked by flow control triggers a STREAM_DATA_BLOCKED frame",
 			packetType1RTT, debugFrameStreamDataBlocked{
@@ -45,15 +107,14 @@ func TestStreamWriteBlockedByStreamFlowControl(t *testing.T) {
 				max: 0,
 			})
 
-		// Blocking write waiting for flow control.
-		w := runAsync(tc, func(ctx context.Context) (int, error) {
-			return s.WriteContext(ctx, want)
-		})
-		tc.wantFrame("second blocked write triggers another STREAM_DATA_BLOCKED",
-			packetType1RTT, debugFrameStreamDataBlocked{
-				id:  s.id,
-				max: 0,
-			})
+		// Write more data.
+		_, err = s.WriteContext(ctx, want[1:])
+		if err != nil {
+			t.Fatalf("write with available output buffer: unexpected error: %v", err)
+		}
+		tc.wantIdle("adding more blocked data does not trigger another STREAM_DATA_BLOCKED")
+
+		// Provide some flow control window.
 		tc.writeFrames(packetType1RTT, debugFrameMaxStreamData{
 			id:  s.id,
 			max: 4,
@@ -69,6 +130,7 @@ func TestStreamWriteBlockedByStreamFlowControl(t *testing.T) {
 				data: want[:4],
 			})
 
+		// Provide more flow control window.
 		tc.writeFrames(packetType1RTT, debugFrameMaxStreamData{
 			id:  s.id,
 			max: int64(len(want)),
@@ -79,10 +141,6 @@ func TestStreamWriteBlockedByStreamFlowControl(t *testing.T) {
 				off:  4,
 				data: want[4:],
 			})
-		n, err := w.result()
-		if n != len(want) || err != nil {
-			t.Errorf("Write() = %v, %v; want %v, nil", n, err, len(want))
-		}
 	})
 }
 
@@ -169,7 +227,7 @@ func TestStreamWriteBlockedByWriteBufferLimit(t *testing.T) {
 			p.initialMaxStreamDataBidiRemote = 1 << 20
 			p.initialMaxStreamDataUni = 1 << 20
 		}, func(c *Config) {
-			c.StreamWriteBufferSize = maxWriteBuffer
+			c.MaxStreamWriteBufferSize = maxWriteBuffer
 		})
 		tc.handshake()
 		tc.ignoreFrame(frameTypeAck)
@@ -391,7 +449,7 @@ func TestStreamReceiveExtendsStreamWindow(t *testing.T) {
 		const maxWindowSize = 20
 		ctx := canceledContext()
 		tc := newTestConn(t, serverSide, func(c *Config) {
-			c.StreamReadBufferSize = maxWindowSize
+			c.MaxStreamReadBufferSize = maxWindowSize
 		})
 		tc.handshake()
 		tc.ignoreFrame(frameTypeAck)
@@ -448,7 +506,7 @@ func TestStreamReceiveViolatesStreamDataLimit(t *testing.T) {
 			size: 2,
 		}} {
 			tc := newTestConn(t, serverSide, func(c *Config) {
-				c.StreamReadBufferSize = maxStreamData
+				c.MaxStreamReadBufferSize = maxStreamData
 			})
 			tc.handshake()
 			tc.ignoreFrame(frameTypeAck)
@@ -473,7 +531,7 @@ func TestStreamReceiveDuplicateDataDoesNotViolateLimits(t *testing.T) {
 		const maxData = 10
 		tc := newTestConn(t, serverSide, func(c *Config) {
 			// TODO: Add connection-level maximum data here as well.
-			c.StreamReadBufferSize = maxData
+			c.MaxStreamReadBufferSize = maxData
 		})
 		tc.handshake()
 		tc.ignoreFrame(frameTypeAck)
@@ -557,7 +615,7 @@ func TestStreamFinalSizePastMaxStreamData(t *testing.T) {
 	finalSizeTest(t, errFlowControl, func(tc *testConn, sid streamID) (finalSize int64) {
 		return 11
 	}, func(c *Config) {
-		c.StreamReadBufferSize = 10
+		c.MaxStreamReadBufferSize = 10
 	})
 }
 
@@ -649,7 +707,7 @@ func TestStreamReceiveUnblocksReader(t *testing.T) {
 // to the conn and expects a STREAM_STATE_ERROR.
 func testStreamSendFrameInvalidState(t *testing.T, f func(sid streamID) debugFrame) {
 	testSides(t, "stream_not_created", func(t *testing.T, side connSide) {
-		tc := newTestConn(t, side)
+		tc := newTestConn(t, side, permissiveTransportParameters)
 		tc.handshake()
 		tc.writeFrames(packetType1RTT, f(newStreamID(side, bidiStream, 0)))
 		tc.wantFrame("frame for local stream which has not been created",
@@ -659,7 +717,7 @@ func testStreamSendFrameInvalidState(t *testing.T, f func(sid streamID) debugFra
 	})
 	testSides(t, "uni_stream", func(t *testing.T, side connSide) {
 		ctx := canceledContext()
-		tc := newTestConn(t, side)
+		tc := newTestConn(t, side, permissiveTransportParameters)
 		tc.handshake()
 		sid := newStreamID(side, uniStream, 0)
 		s, err := tc.conn.NewSendOnlyStream(ctx)
@@ -796,7 +854,7 @@ func TestStreamOffsetTooLarge(t *testing.T) {
 }
 
 func TestStreamReadFromWriteOnlyStream(t *testing.T) {
-	_, s := newTestConnAndLocalStream(t, serverSide, uniStream)
+	_, s := newTestConnAndLocalStream(t, serverSide, uniStream, permissiveTransportParameters)
 	buf := make([]byte, 10)
 	wantErr := "read from write-only stream"
 	if n, err := s.Read(buf); err == nil || !strings.Contains(err.Error(), wantErr) {
@@ -868,16 +926,15 @@ func TestStreamWriteToClosedStream(t *testing.T) {
 }
 
 func TestStreamResetBlockedStream(t *testing.T) {
-	tc, s := newTestConnAndLocalStream(t, serverSide, bidiStream, func(p *transportParameters) {
-		p.initialMaxStreamsBidi = 1
-		p.initialMaxData = 1 << 20
-		p.initialMaxStreamDataBidiRemote = 4
-	})
+	tc, s := newTestConnAndLocalStream(t, serverSide, bidiStream, permissiveTransportParameters,
+		func(c *Config) {
+			c.MaxStreamWriteBufferSize = 4
+		})
 	tc.ignoreFrame(frameTypeStreamDataBlocked)
 	writing := runAsync(tc, func(ctx context.Context) (int, error) {
 		return s.WriteContext(ctx, []byte{0, 1, 2, 3, 4, 5, 6, 7})
 	})
-	tc.wantFrame("stream writes data until blocked by flow control",
+	tc.wantFrame("stream writes data until write buffer fills",
 		packetType1RTT, debugFrameStream{
 			id:   s.id,
 			off:  0,
@@ -894,11 +951,8 @@ func TestStreamResetBlockedStream(t *testing.T) {
 	if n, err := writing.result(); n != 4 || !strings.Contains(err.Error(), wantErr) {
 		t.Errorf("s.Write() interrupted by Reset: %v, %q; want 4, %q", n, err, wantErr)
 	}
-	tc.writeFrames(packetType1RTT, debugFrameMaxStreamData{
-		id:  s.id,
-		max: 1 << 20,
-	})
-	tc.wantIdle("flow control is available, but stream has been reset")
+	tc.writeAckForAll()
+	tc.wantIdle("buffer space is available, but stream has been reset")
 	s.Reset(100)
 	tc.wantIdle("resetting stream a second time has no effect")
 	if n, err := s.Write([]byte{}); err == nil || !strings.Contains(err.Error(), wantErr) {
@@ -1040,6 +1094,44 @@ func TestStreamCloseUnblocked(t *testing.T) {
 	}
 }
 
+func TestStreamCloseWriteWhenBlockedByStreamFlowControl(t *testing.T) {
+	ctx := canceledContext()
+	tc, s := newTestConnAndLocalStream(t, serverSide, uniStream, permissiveTransportParameters,
+		func(p *transportParameters) {
+			//p.initialMaxData = 0
+			p.initialMaxStreamDataUni = 0
+		})
+	tc.ignoreFrame(frameTypeStreamDataBlocked)
+	if _, err := s.WriteContext(ctx, []byte{0, 1}); err != nil {
+		t.Fatalf("s.Write = %v", err)
+	}
+	s.CloseWrite()
+	tc.wantIdle("stream write is blocked by flow control")
+
+	tc.writeFrames(packetType1RTT, debugFrameMaxStreamData{
+		id:  s.id,
+		max: 1,
+	})
+	tc.wantFrame("send data up to flow control limit",
+		packetType1RTT, debugFrameStream{
+			id:   s.id,
+			data: []byte{0},
+		})
+	tc.wantIdle("stream write is again blocked by flow control")
+
+	tc.writeFrames(packetType1RTT, debugFrameMaxStreamData{
+		id:  s.id,
+		max: 2,
+	})
+	tc.wantFrame("send remaining data and FIN",
+		packetType1RTT, debugFrameStream{
+			id:   s.id,
+			off:  1,
+			data: []byte{1},
+			fin:  true,
+		})
+}
+
 func TestStreamPeerResetsWithUnreadAndUnsentData(t *testing.T) {
 	testStreamTypes(t, "", func(t *testing.T, styp streamType) {
 		ctx := canceledContext()
@@ -1112,7 +1204,7 @@ func TestStreamPeerResetFollowedByData(t *testing.T) {
 }
 
 func TestStreamResetInvalidCode(t *testing.T) {
-	tc, s := newTestConnAndLocalStream(t, serverSide, uniStream)
+	tc, s := newTestConnAndLocalStream(t, serverSide, uniStream, permissiveTransportParameters)
 	s.Reset(1 << 62)
 	tc.wantFrame("reset with invalid code sends a RESET_STREAM anyway",
 		packetType1RTT, debugFrameResetStream{
@@ -1163,6 +1255,23 @@ func TestStreamPeerStopSendingForActiveStream(t *testing.T) {
 	})
 }
 
+func TestStreamReceiveDataBlocked(t *testing.T) {
+	tc := newTestConn(t, serverSide, permissiveTransportParameters)
+	tc.handshake()
+	tc.ignoreFrame(frameTypeAck)
+
+	// We don't do anything with these frames,
+	// but should accept them if the peer sends one.
+	tc.writeFrames(packetType1RTT, debugFrameStreamDataBlocked{
+		id:  newStreamID(clientSide, bidiStream, 0),
+		max: 100,
+	})
+	tc.writeFrames(packetType1RTT, debugFrameDataBlocked{
+		max: 100,
+	})
+	tc.wantIdle("no response to STREAM_DATA_BLOCKED and DATA_BLOCKED")
+}
+
 type streamSide string
 
 const (
@@ -1216,3 +1325,11 @@ func permissiveTransportParameters(p *transportParameters) {
 	p.initialMaxStreamDataBidiLocal = maxVarint
 	p.initialMaxStreamDataUni = maxVarint
 }
+
+func makeTestData(n int) []byte {
+	b := make([]byte, n)
+	for i := 0; i < n; i++ {
+		b[i] = byte(i)
+	}
+	return b
+}
